#!/usr/bin/perl

# ToDo: Hanja sorting! use Unihan.txt???
# No! use instead, \index{Hangul@Hanja}

use 5.8.0;
use strict;
use Getopt::Std;

if(@ARGV == 0) {
  print "Usage: $0 [makeindex options]\n";
  exit 1;
}

my $ispace = "\n  \\indexspace\n\n";
my @cjamo_to_jamo= qw/
1100 1101 11aa 1102 11ac 11ad 1103 1104
1105 11b0 11b1 11b2 11b3 11b4 11b5 111a
1106 1107 1108 1121 1109 110a 110b 110c
110d 110e 110f 1110 1111 1112 1161 1162
1163 1164 1165 1166 1167 1168 1169 116a
116b 116c 116d 116e 116f 1170 1171 1172
1173 1174 1175 1160 1114 1115 11c7 11c8
11cc 11ce 11d3 11d7 11d9 111c 11dd 11df
111d 111e 1120 1122 1123 1127 1129 112b
112c 112d 112e 112f 1132 1136 1140 1147
114c 11f1 11f2 1157 1158 1159 1184 1185
1188 1191 1192 1194 119e 11a1
/;

my %args;
getopt('cgilo:p:qrs:t:LT',\%args);

my @IDXfiles = @ARGV;
@IDXfiles or die "idx0 is not defined!\n";

my $istfile;
if($args{'s'}) {
  $istfile = $args{'s'};
  $istfile =~ s/\.ist$//; $istfile .= ".ist";
}

my $indfile = $args{o} || $IDXfiles[0];
$indfile =~ s/\.(ind|idx)$//; $indfile .= ".ind";

my $logfile = $args{t} || $IDXfiles[0];
$logfile =~ s/\.(ilg|idx)$//; $logfile .= ".ilg";

foreach my $file (@IDXfiles) {
  $file =~ s/\.idx$//; $file .= ".idx";
}

my $makeindexoption = "-i -o $indfile -t $logfile";
foreach ('c','g','l','q','r','L','T') {
  $makeindexoption .= " -$_" if $args{$_};
}
$makeindexoption .= " -s $istfile" if $args{'s'};
$makeindexoption .= " -p $args{p}" if $args{p};

my @idxarr;
foreach my $file(@IDXfiles) {
  open IDX,$file or die "$file : $!\n";
  while (<IDX>) {
    if(/\\indexentry\{(.+)\}\{(.+)\}/) {
      my ($pre,$post) = ($1,$2);
      if($pre =~ /(.+?[^\"])\@(.+)/) {
      	my($prepre,$prepost) = ($1,$2);
      	$prepre =~ s/\^\^\^\^([0-9a-f]{4})/topronun($1)/ge;
      	push @idxarr,"\\indexentry{$prepre\@$prepost}{$post}\n";
      }
      elsif($pre =~ /(.+?[^\"])\|(.+)/) {
      	my($prepre,$prepost) = ($1,$2);
      	my $preprepre = $prepre =~ /(.*?)\!/ ? $1 : $prepre;
      	$preprepre =~ s/\^\^\^\^([0-9a-f]{4})/topronun($1)/ge;
      	push @idxarr,"\\indexentry{$preprepre\@$prepre|$prepost}{$post}\n";
      }
      else {
      	my $prepre = $pre =~ /(.+?)\!/ ? $1 : $pre;
      	$prepre =~ s/\^\^\^\^([0-9a-f]{4})/topronun($1)/ge;
      	push @idxarr,"\\indexentry{$prepre\@$pre}{$post}\n";
      }
    }
  }
  close IDX;
}

print "makeindex $makeindexoption\n";
my $pid = open(MAKE,"| makeindex $makeindexoption") or die "Cannot fork: $!\n";
print MAKE @idxarr;
close MAKE;
$? >> 8 and die "makeindex failed!\n";

my (@Latin,@NonLatin);
open IND,$indfile or die "$indfile : $!\n";
my $spacefound = 0;
while(<IND>) {
  if(/^\\(begin|end)\{theindex\}/) {next;}
  elsif($spacefound == 0 and /\\indexspace/) { $spacefound = 1; }
  elsif($spacefound == 0 and /item/) { push @NonLatin,$_; }
  elsif($spacefound != 0) { push @Latin,$_; }
}
close IND;

my($heading_prefix,$heading_suffix,$heading_flag);
if($istfile) {
  $istfile = `kpsewhich $istfile`;
  chomp $istfile;
  open IST,$istfile or die "$istfile : $!\n";
  while (<IST>) {
    if(/headings_flag\s+(.+)/){$heading_flag=$1;}
    elsif(/heading_prefix\s+"(.+)"/){$heading_prefix=$1;}
    elsif(/heading_suffix\s+"(.+)"/){$heading_suffix=$1;}
  }
  close IST;
  $heading_prefix =~ s/\\\\/\\/g;
  $heading_suffix =~ s/\\\\/\\/g;
}

my $pre = "";
my (@Hangul,@Symbol,$hangul);

foreach my $i (0 .. $#NonLatin) {
  next if($heading_flag and
	  $NonLatin[$i]=~/^\Q${heading_prefix}Symbols$heading_suffix/i);
  if($NonLatin[$i] =~ /\\item /) {
    my $first;

    my $j = identifyJ($NonLatin[$i]);
    if($idxarr[$j] =~ /.+?\{\^\^\^\^([0-9a-f]{4})/) { $first = $1; }

    if($first and hex($first) >= 0x1100 and hex($first) <= 0x115f ) {
      $hangul = 1;
      if($pre eq $first) { push @Hangul,$NonLatin[$i]; }
      else {
	my $prefix="";
	if($heading_flag){ $prefix = gethangulprefix($first);}
	push @Hangul,$ispace,$prefix,$NonLatin[$i];
	$pre = $first;
      }
    }
    else {
      $hangul = 0;
      push @Symbol,$NonLatin[$i];
    }
  }
  else {
    if($hangul) { push @Hangul,$NonLatin[$i]; }
    else { push @Symbol,$NonLatin[$i]; }
  }
}

if(@Symbol) {
  $heading_flag and
    unshift @Symbol,$heading_prefix."^^^^ae30^^^^d638".$heading_suffix."\n";
  push @Symbol,$ispace;
}

foreach(@Symbol,@Latin,@Hangul) {
  s/\^\^\^\^([0-9a-f]{4})/chr hex $1/eg;
  s/\^\^([0-9a-f]{2})/chr hex $1/eg;
}

open IND,">$indfile" or die "$indfile : $!\n";
binmode(IND, ":utf8");
print IND "\\ocp\\OCPinutf=inutf8
\\InputTranslation currentfile \\OCPinutf

\\begin{theindex}
",@Symbol,@Latin,@Hangul,"
\\end{theindex}\n";
close IND;


sub topronun {
  my $hexa = hex shift;

  if($hexa >= 0xac00 and $hexa <= 0xd7a3) {
    return decompose($hexa);
  }
  elsif($hexa >= 0x3131 and $hexa <= 0x318e) {
    my $jamo = $cjamo_to_jamo[$hexa - 0x3131];
    if(hex($jamo) >= 0x1100 and hex($jamo) <= 0x115f) {
      return "^^^^$jamo^^^^1160";
    } elsif(hex($jamo) >= 0x1160 and hex($jamo) <= 0x11a2) {
      return "^^^^115f^^^^$jamo";
    } else {
      return "^^^^115f^^^^1160^^^^$jamo";
    }
  }
  else { return "^^^^".sprintf("%04x",$hexa); }
}

sub decompose {
  my $code = shift;
  ### http://www.unicode.org/reports/tr15/ 
  my ($Sbase,$Lbase,$Vbase,$Tbase) = (0xac00,0x1100,0x1161,0x11a7);
  my ($Lcount,$Vcount,$Tcount) = (19,21,28);
  my $Ncount = $Vcount * $Tcount;   # 588
  my $Scount = $Lcount * $Ncount;   # 11172
  my $Sindex = $code - $Sbase;
  my $cho = $Lbase + $Sindex / $Ncount;
  my $jung = $Vbase + ($Sindex % $Ncount) / $Tcount;
  my $jong = $Tbase + $Sindex % $Tcount;
  my $result = "^^^^".sprintf("%04x",$cho)."^^^^".sprintf("%04x",$jung);
  $result .= "^^^^".sprintf("%04x",$jong) if $jong != $Tbase;
  return $result;
}

sub identifyJ {
  my $cnt = shift;
  $cnt =~ s/.*\\item //;
  $cnt =~ s/,.*//;

  my $i = 0;
  for ($i=0;$i<@idxarr;$i++) {
    my $idx = $idxarr[$i];
    $idx=~s/.*\\indexentry\{//;
    $idx=~s/\}.+//;
    last if($idx =~ /\Q$cnt\E/);
  }
  return $i;
}

sub gethangulprefix {
  my $first = shift;
  my ($cjamo,$i);
  for ($i=0;$i<@cjamo_to_jamo;$i++){
    if(hex($cjamo_to_jamo[$i]) == hex($first)) {
      $cjamo = $i + hex(3131);
      last;
    }
  }
  $cjamo = sprintf("%04x",$cjamo);
  return $heading_prefix."^^^^".$cjamo.$heading_suffix."\n";
}
